import numpy as np 
from scipy.integrate import solve_ivp
try: import jax
except: pass


class NewtonRaphson:
    def __new__(cls, Model, U0 = None, tol=1.0e-3, loss=None, verbose=False,disp_dof=0):
        nf = Model.nf
        nt = Model.nt
        if U0 is None: U0 = np.zeros(nt)
        
        # create pure function Pr = Pr_U( U ) from function Pr = Pr_vector(U, Model)
        Pr_U = partial(emx.Pr_vector, Model )

        # use 'forward mode' algorithmic differentiation 
        # to obtain jacobian of Pr_U(U), Kt(U)
        Jac = jax.jacfwd(Pr_U)
        Kt = lambda Ui: np.squeeze(Jac(Ui))[:nf,:nf]
    
        @jax.jit
        def _dPu_(U, Pf):
            Pr = Pr_U(U)[:nf]
            Pu = Pf - Pr
            Kf = Kt(U)
            DUf = np.squeeze(np.linalg.solve(Kf,Pu))

            # reshaping
            DU = np.pad(DUf, (0, Model.nr), 'constant') 

            # update displacement
            Unew = U + DU
            return Unew, Pu
        
        
        def solve(Pf, verbose=False, maxiter=100):
            r = tol+1
            count = 1
            U = U0
            while jax.lax.gt(r , tol ):
                U, Pu = _dPu_(U,Pf)
                
                r = np.linalg.norm(Pu)

                if jax.lax.gt(count , maxiter): break
                if verbose: print('{}: {:.3f} {:.3f}'.format(count, r, U[disp_dof,...]))
                count+=1
            return U
            
        return solve

def ElcentroRHA(zeta, omega):
    g=386.4
    ground_motion = np.genfromtxt('ElcentroNS.csv', delimiter=',')[:750,:]
    ground_motion[0,0]=0.0
    time = ground_motion[:,0]
    t_span = (time[0], time[-1])
    u0 = (0.0, 0.0)

    def EOM(t, y):
        i = np.argmin(np.abs(time - t))
        D, dDdt = y
        dydt = [dDdt, -ground_motion[i,1]*g-omega**2*D-2*zeta*omega*dDdt]
        return dydt

    rtol=atol=1e-5
    sol = solve_ivp(EOM, y0=u0, t_span=t_span, args=(zeta, omega), rtol=rtol, atol=atol)
    D = sol.y[0,:]
    t = sol.t
    return t, D
   
def ExcitationInterpolation(p, Dt, T, u0, v0, k, Tn, zeta=0):
    n  = int(T/Dt) + 1
    c  = zeta*k*Tn/np.pi
    m  = k/(2*np.pi/Tn)**2
    wn = 2*np.pi/Tn
    wd = wn*np.sqrt(1 - zeta**2)
    
    # Initialize arrays
    u = np.zeros((n, 1))
    u[0] = v0
    vel = np.zeros((n,1))
    vel[0] = v0
    
    A = np.exp(-zeta*wn*Dt)*(zeta/np.sqrt(1-zeta**2)*np.sin(wd*Dt)+np.cos(wd*Dt))
    B = np.exp(-zeta*wn*Dt)*(1/wd*np.sin(wd*Dt))
    C = 1/k*(2*zeta/(wn*Dt)+np.exp(-zeta*wn*Dt)*(((1-2*zeta**2)/(wd*Dt)-zeta/np.sqrt(1-zeta**2))*np.sin(wd*Dt)-(1+2*zeta/(wn*Dt))*np.cos(wd*Dt)))
    D = 1/k*(1-2*zeta/(wn*Dt)+np.exp(-zeta*wn*Dt)*(2*zeta**2-1)/(wn*Dt)*np.sin(wd*Dt)+2*zeta/(wn*Dt)*np.cos(wd*Dt))
    A1 = -np.exp(-zeta*wn*Dt)*(wn/np.sqrt(1-zeta**2)*np.sin(wd*Dt))
    B1 =  np.exp(-zeta*wn*Dt)*(np.cos(wd*Dt)-zeta/np.sqrt(1-zeta**2)*np.sin(wd*Dt))
    C1 = 1/k*(-1/Dt+np.exp(-zeta*wn*Dt)*((wn/(np.sqrt(1-zeta**2))+zeta/(Dt*np.sqrt(1-zeta**2)))*np.sin(wd*Dt)+1/Dt*np.cos(wd*Dt)))
    D1 = 1/(k*Dt)*(1-np.exp(-zeta*wn*Dt)*(zeta/np.sqrt(1-zeta**2)*np.sin(wd*Dt)+np.cos(wd*Dt)))
    
 #     print(A,B,C,D,A1,B1,C1,D1)
    for i in range(n-1):
        u[i+1] = A*u[i]+B*vel[i]+C*p(Dt*i)+D*p(Dt*(i+1))
        vel[i+1] = A1*u[i]+B1*vel[i]+C1*p(Dt*i)+D1*p(Dt*(i+1))
    # u due to free vibration
    return u, vel

def CentralDifference(p, Dt, T, u0, v0, k, Tn, zeta=0):
    n    = int(T/Dt)+1
    c    = zeta*k*Tn/np.pi
    m    = k/(2*np.pi/Tn)**2
    
    a0 = (p(0.0)-c*v0-k*u0)/m
    k_hat = m/Dt**2+c/(2*Dt)
    alpha = m/Dt**2-c/(2*Dt)
    beta = k - 2*m/Dt**2
    ui = np.zeros(n)
    uip1 = np.zeros(n)
    uim1 = np.zeros(n)
    p_hat = np.zeros(n)
    
    for i in range(1,n):
        t = (i-1)*Dt
        if i==1:
            ui[i]=u0
            uim1[i] = u0-(Dt)*v0+0.5*Dt**2*a0
        else:
            ui[i] = uip1[i-1]
            uim1[i] = ui[i-1]
            
        p_hat[i] =  p(t) - alpha*uim1[i] - beta*ui[i]
        uip1[i] = p_hat[i]/k_hat
    
    return uip1

def Newmark(p, Dt, T, u0, v0, k, Tn, zeta=0):
    n = int(T/Dt)+1
    c = zeta*k*Tn/np.pi
    m = k/(2*np.pi/Tn)**2
    
    u = np.zeros(n)
    v = np.zeros(n)
    a = np.zeros(n)
    p_hat = np.zeros(n)
    
    u[0] = u0
    v[0] = v0
    a[0] = (p(0.0) - c*v0 - k*u0)/m
    a1 = 4/Dt**2*m + 2/Dt*c
    a2 = 4/Dt*m + c
    a3 = m
    k_hat = k + a1
    
    for i in range(1, n-1):
        t = (i-1)*Dt
        p_hat[i+1] = p(t) +a1*u[i]+a2*v[i]+a3*a[i]
        u[i+1] = p_hat[i+1]/k_hat
        v[i+1] = 2/Dt*(u[i+1]-u[i])-v[i]
        a[i+1] = 4/Dt**2*(u[i+1]-u[i])-4/Dt*v[i]-a[i]

    return u, v, a

def Herrmann_Peterson(E, tau, epsilon, tot_time, N):
    dt = tot_time/N  
    time = np.zeros((N,1))
    sigma = np.zeros((N,1))
    h1 = np.zeros((N,1))
    h2 = np.zeros((N,1))
    h3 = np.zeros((N,1))

    for i in range(1, N):
        time[i] = time[i-1]+dt

        de = epsilon(time[i]) - epsilon(time[i-1])
        h1[i] = np.exp(-dt/tau[1])*h1[i-1] + E[1]*(tau[1]/dt)*(1- np.exp(-dt/tau[1]))*de 
        h2[i] = np.exp(-dt/tau[2])*h2[i-1] + E[2]*(tau[2]/dt)*(1- np.exp(-dt/tau[2]))*de 
        h3[i] = np.exp(-dt/tau[3])*h2[i-1] + E[3]*(tau[3]/dt)*(1- np.exp(-dt/tau[3]))*de 
        sigma[i] = E[0]*epsilon(time[i]) + h1[i] + h2[i] + h3[i] 
        
    return time, sigma

def _armijo_haukaas_init(b0: float=0.2, b: float = 0.5, m: int =4, a: float=0.5):
    
    def _armijo_haukaas():
        pass 
    try: 
        _armijo_haukaas = jax.jit(_armijo_haukaas)
        print('Jit Successful')
    except: pass
    return _armijo_haukaas

def _merit_init():
    pass

class iHLRF:
    def __init__(self, f, gradf, u0, loss=None, 
                    tol1=1.0e-4, tol2=1.0e-4, maxiter=20, maxiter_step=10,
                    step_opt={'c':10.},**kwargs):
        self.u0 = u0
        self.f = f 
        self.gradf = gradf
        self.tol1 = tol1
        self.tol2 = tol2
        self.maxiter = maxiter
        self.maxiter_step = maxiter_step
        self.loss = loss
        self.c = step_opt['c']
        if loss is None: 
            self.loss = np.linalg.norm

        # self.init()

    def init(self,verbose=False):
        if verbose:
            print('\niHL-RF Algorithm (Zhang and Der Kiureghian 1995)***************',
                  '\nInitialize iHL-RF: \n',
                  'u0: ', np.around(self.u0.T,4), '\n')

        self.G0 = self.f(self.u0) # scale parameter
        if verbose: print('G0: ',np.around(self.G0,4),'\n')

        self.ui = self.u0[None,:].T

        self.Gi = self.G0
        if verbose: print('Gi: ', np.around(self.Gi,4), '\n',)

        self.GradGi = self.gradf(self.ui[:,0].T)
        # if verbose: print('GradGi: ', np.around(self.GradGi,4), '\n',)
        if verbose: print('GradGi: ', self.GradGi, '\n',)

        self.alphai = -(self.GradGi / np.linalg.norm(self.GradGi))[None,:]
        self.count = 0
        self.res1 = abs(self.Gi/self.G0)
        self.res2 = self.loss(self.ui - (self.alphai@self.ui)*self.alphai.T )

        if verbose:
            print(
                # '\niHL-RF Algorithm (Zhang and Der Kiureghian 1995)***************',
                #   '\nInitialize iHL-RF: \n',
                #   'u0: ', np.around(self.u0.T,4), '\n',
                #   'G0: ',np.around(self.G0,4),'\n'
                #   'ui: ', np.around(self.ui.T,4), '\n',
                #   'Gi: ', np.around(self.Gi,4), '\n',
                #   'GradGi: ', np.around(self.GradGi,4),'\n',
                  'alphai: ', np.around(self.alphai,4),'\n',
                  'res1: ' , np.around(self.res1,4),'\n',
                #   'res2: ' , np.around(self.res2,4),'\n',
                  'res2: ' , self.res2,'\n',)
    def merit():
        pass

    def incr(self,method='adk', verbose=False):
        if method == 'basic':
            self.ui1 = self.ui + self.lamda * self.di
            return self.ui1 
        # c = 10.0
        self.ci = self.loss(self.ui) / np.linalg.norm(self.GradGi) + self.c
        self.mi = 0.5*self.loss(self.ui)**2 + self.ci*abs(self.Gi)
        self.mi1 = 0.5*self.loss(self.ui1)**2 + self.ci*abs(self.Gi1)

        self.count_step = 0
        while (self.mi1 >= self.mi) :
            self.lamda = self.lamda/2
            if verbose: print('lamda:', self.lamda)
            self.ui1 = self.ui + self.lamda * self.di
            if verbose: print('ui1: ',self.ui1)
            self.Gi1 = self.f(self.ui1[:,0].T)
            self.mi1 = 0.5*np.linalg.norm(self.ui1)**2 + self.ci*abs(self.Gi1)
            self.count_step += 1
            if (self.count_step >= self.maxiter_step): break
        return self.ui1

    
    
    def dirn(self):
        self.di = (self.Gi/np.linalg.norm(self.GradGi) + self.alphai@self.ui)*self.alphai.T - self.ui
        return self.di

    def step(self,verbose=False,basic=False):
        # self.di = (self.Gi/np.linalg.norm(self.GradGi) + self.alphai@self.ui)*self.alphai.T - self.ui
        di = self.dirn()
        if verbose: print('di: ',self.di)

        self.lamda = 1.0
        # self.lamda = 0.05

        self.ui1 = self.ui + self.lamda * di
        if verbose: print('ui1: ',self.ui1,self.ui1[:,0].T)

        self.Gi1 = self.f(self.ui1[:,0].T)

        self.incr(basic)

        self.ui = self.ui1  
        self.Gi = self.f(self.ui[:,0].T)
        self.GradGi = self.gradf(self.ui[:,0].T)
        self.alphai = -(self.GradGi / np.linalg.norm(self.GradGi))[None,:]

        self.res1 = abs(self.Gi/self.G0)
        self.res2 = self.loss(self.ui - (self.alphai@self.ui)*self.alphai.T)
        self.count += 1


        if verbose: print('\niHL-RF step: {}'.format(self.count))

        if verbose:
            print('ui: ',       np.around(self.ui,4), '\n',
                  'Gi: ',       np.around(self.Gi,4), '\n',
                  'GradGi: ', np.around(self.GradGi,4),'\n',
                  'alphai: ', np.around(self.alphai,4),'\n',
                  'res1: ' , np.around(self.res1,4),'\n',
                  'res2: ' , np.around(self.res2,4),'\n',)
        return self.ui
    
    def run(self,verbose=False, steps=None):
        self.init(verbose=verbose)
        if steps is not None: self.maxiter = steps
        while not(self.res1 < self.tol1 and self.res2 < self.tol2):
            self.step(verbose=verbose)

            if (self.count > self.maxiter): break

        return self.ui



import numpy as np


def GMP(fy,E,Eh,r):
    # @jax.jit
    def sig(eps):
        xi = eps/(fy/E)
        b = Eh/E
        return fy*(xi*b+(1-b)*xi/(1+abs(xi)**r)**(1/r))

    # @jax.jit
    def Et(eps):
        xi = eps/(fy/E)
        b = Eh/E
        return E*(b+(1-b)/(1+abs(xi)**r)**(1+1/r))
    
    return sig, Et

def BiLinOld(epsy=None,E=None,fy=None,Eh=0.0,r=None,activ=None, MatData=True):
    if epsy is None: epsy = fy/E 
    if fy is None: fy = epsy*E
    epspmax = 1.0

    # @jax.jit
    def sig(eps):
        return min(eps,epsy)*E + max(0.0, min(epspmax,eps-epsy))*Eh

    # @jax.jit
    def Et(eps):
        return (E,Eh)[int((eps>=epsy))]

    if MatData:
        MatData = {'sig':sig,'Et':Et,'epsy':epsy,'fy':fy}
        return sig, Et, MatData

    return sig, Et

def BiLin(epsy=None,E=None,fy=None,Eh=0.0,r=None,activ=None, MatData=True):
    if epsy is None: epsy = fy/E 
    if fy is None: fy = epsy*E
    epspmax = 1.0

    def sig(eps):
        ex = np.clip(eps,-epsy,epsy)
        sgn = np.sign(ex)
        exp = np.clip((eps-ex)*sgn, -1.0,  1.0)*sgn
        return ex*E + exp*Eh

    def Et(eps):
        return (E,Eh)[int(abs(eps)>=epsy)]

    if MatData:
        MatData = {'sig':sig,'Et':Et,'epsy':epsy,'fy':fy}
        return sig, Et, MatData

    return sig, Et

def Mander88(e_c, f_c, f_cc, f_emin, confined=True):
    """Returns an array of values along the stress-strain curve for a 
    specified concrete section according to the relation proposed by 
    Mander et al. in 1988.

    :param f_c: Specified 28-day concrete compression strength
    :type f_c: float
    :param f_cc: Confined concrete compressive strength
    :type f_cc: float
    :param f_emin: Lesser of the effective confinement stresses
    :type f_emin: float

    :return: a list of strain values, and a list of stress values
    :rtype: : list
    """

    C = 0.85
    epsilon_0 = 0.002
    E_c = aci.e19_2_2_1a(C*f_c)
    
    if confined:
        epsilon_cc = epsilon_0*(1+5*(f_cc/(C*f_c)-1)) # eq 4.16, page 102
        epsilon_u = 0.004+0.25*f_emin/(C*f_c)
    else:
        epsilon_cc = epsilon_0
        epsilon_u = 0.004


    # epsilon_c = np.linspace(0, epsilon_u, 100)
    epsilon_c = e_c
    r = E_c/(E_c-f_cc/epsilon_cc)
    
    x = epsilon_c/epsilon_cc
    fc = f_cc*x*r/(r-1+x**r)
    return fc

def Mander84(e_s, strainHardening=True): # page 46, Moehle 2014
    """Returns an array of values along the stress-strain curve for steel
    reinforcement according to the relation proposed by 
    Mander et al. in 1984.

    """
    Grade = 'A706'
    Gr = {
        'A706': {
            'fy' :69000,
            'fsu':95000,
            'Esh':1000000, # fig 5.9
        },
    }
    E = steel[Grade]['E']
    f_ey  = steel[Grade]['f_ey']
    f_su = Gr[Grade]['fsu']
    E_sh = Gr[Grade]['Esh']
    epsilon_0 = 0.002
    epsilon_su = 0.14

    # epsilon_s = np.linspace(0,0.023,100)
    # epsilon_sh = 2*epsilon_0 # fig 5.7
    # epsilon_sh = 10*f_ey/E
    epsilon_sh = 0.016
    if e_s < f_ey/E:
        f_s = E*e_s
    elif e_s < epsilon_sh:
        f_s = f_ey
    elif e_s > epsilon_sh:
        f_s = f_ey
    elif e_s > epsilon_sh and strainHardening == True:
        f_s = f_su+(f_ey-f_su)*((epsilon_su-e_s)/(epsilon_su-epsilon_sh))**(E_sh*((epsilon_su-epsilon_sh)/(f_su-f_ey)))
    
    return f_s




import numpy as np
import scipy.linalg
# import sympy as sp
import matplotlib.pyplot as plt
# import pandas as pd
import emma.elements 
import sympy as sp


def transfer_vars(item1, item2):
    for key, value in item1.__dict__.items():
        item2.__dict__[key] = value

class Structural_Vector(np.ndarray):
    column_data = ["vector"]
    row_data = None
    subs = [None]
    def __new__(cls, mat):
        mat = mat
        return np.asarray(mat).view(cls)

    def __init__(self, mat):
        self.tag = None

    def _repr_html_(self):
        try:
            out = self.df.to_html()
            return out
        except: 
            return self

    def __add__(self, v1):
        if type(v1) is type(self):
            out = np.add(self, v1).view(type(self))
            transfer_vars(self,out)
        return out
    
    def get(self, component):
        row = self.row_data.index(component)
        return self[row]

    def set_item(self, key, value):
        idx = self.row_data.index(key)
        self[idx] = value

    def rows(self, component):
        rows = [self.row_data.index(item) for item in component]
        newV = self[rows]
        newV.row_data = list(np.array(self.row_data)[rows])
        newV.model = self.model
        return newV

    @property
    def df(self):
        row_data = ['$'+tag+'$' for tag in self.row_data]
        header = '$'+self.tag+'_{{'
        try:
            for sub in self.subs:
                header += subscripts[sub]
        except: pass
        print
        header += '}}$'

        return pd.DataFrame(np.around(self,14), index=row_data, columns=[header])

    @property
    def symb(self):
        var = []
        for eid in self.row_data:
            var.append(sp.symbols(self.tag+eid))
        return sp.Matrix(var)
    
    @property
    def disp(self):
        return sp.Matrix(self)

class Structural_Matrix(np.ndarray):
    column_data = None
    row_data = None
    c_ridx = None # row indexes for use in .c method
    c_cidx = None # column indexes for use in .c method
    tag = None

    def __new__(cls, mat):
        return np.asarray(mat).view(cls)

    def __init__(self, mat):
        self.tag = None

    def __matmul__(self, other):
        if isinstance(other, Structural_Matrix):
            out = np.matmul(self,other).view(Structural_Matrix)
            transfer_vars(self,out)
            out.column_data = other.column_data
        
        elif isinstance(other, Structural_Vector):
            out = np.matmul(self,other).view(Structural_Vector)
            out.row_data = self.row_data
            out.column_data = ['.']
        else:
            out = np.matmul(self,other)
        return out

    def __add__(self, other):
        out = np.add(self,other)
        if (isinstance(other, Structural_Matrix)
            or isinstance(other, Structural_Vector)):
            out.row_data = self.row_data

        return out
    
    def __truediv__(self, other):
        out = np.ndarray.__truediv__(self, other) 
        if (isinstance(other, float) 
            or isinstance(other, Structural_Matrix)
            or isinstance(other, Structural_Vector)):
            out = np.ndarray.__truediv__(self, other).view(Structural_Matrix)
            transfer_vars(self, out)
        else:
            out = np.ndarray.__truediv__(self, other)
        return out
        
    def _repr_html_(self):
        try: 
            df = self.df
            return df.to_html()
        except:
            try:
                return pd.DataFrame(self).to_html()
            except:
                pass

    @property
    def disp(self):
        return sp.Matrix(self)
    
    @property
    def df(self):
        if settings['DATAFRAME_LATEX']:
            row_data = ['$'+tag+'$' for tag in self.row_data]
            column_data = ['$'+tag+'$' for tag in self.column_data]
        else:
            row_data = self.row_data
            column_data = self.column_data
        return pd.DataFrame(np.around(self,14), index=row_data, columns=column_data)

    @property
    def inv(self):
        mat = np.linalg.inv(self)
        transfer_vars(self, mat)
        mat.row_data = self.column_data
        mat.column_data = self.row_data
        return mat
    
    @property
    def rank(self):
        return np.linalg.matrix_rank(self)
    
    @property
    def lns(self):
        return scipy.linalg.null_space(self.T)
    
    @property
    def nls(self):
        return scipy.linalg.null_space(self)
    
    @property
    def ker(self):
        kernel = scipy.linalg.null_space(self) 
        ker = Structural_Matrix(kernel)
        ker.row_data = self.column_data 
        ker.column_data = [str(i+1) for i in range(len(ker[0]))]
        ker.model = self.model
        return ker

    def lu(self):
        
        return scipy.linalg.lu(self)

    @property
    def c(self):
        delcols = self.c_cidx
        delrows = self.c_ridx

        newM = np.delete(self, delrows, axis=0).view(type(self))
        if delcols: newM = np.delete(newM, delcols, axis=1).view(type(self))
        
        transfer_vars(self, newM)

        if delcols: newM.column_data = list(np.delete(self.column_data, delcols))

        newM.row_data = list(np.delete(self.row_data, delrows))
        return newM

    def round(self, num):
        newM = np.around(self, num).view(Structural_Matrix)
        transfer_vars(self,newM)
        return newM

    def remove(self, component):
        """Remove items by looking up column_data/row_data"""
        if type(component) is list:
            for item in component:
                if item in self.column_data: 
                    delcol = self.column_data.index(item)
                    newM = np.delete(self, delcol, axis=1).view(type(self))
                    transfer_vars(self,newM)
                    newM.column_data = list(np.delete(self.column_data, delcol))
                else:
                    delrow = self.row_data.index(item)
                    newM = np.delete(self, delrow, axis=0).view(type(self))
                    transfer_vars(self, newM)
                    # newM.column_data = self.column_data
                    # newM.model = self.model
                    newM.row_data = list(np.delete(self.row_data, delrow))

        else:
            item = component
            if item in self.column_data: 
                delcol = self.column_data.index(item)
                newM = np.delete(self, delcol, axis=1).view(type(self))
                newM.row_data = self.row_data
                newM.model = self.model
                newM.column_data = list(np.delete(self.column_data, delcol))
                # try: newM.rel = self.rel
                # except: pass
            else:
                delrow = self.row_data.index(item)
                newM = np.delete(self, delrow, axis=0).view(type(self))
                newM.column_data = self.column_data
                newM.model = self.model
                newM.row_data = list(np.delete(self.row_data, delrow))
                # try: newM.rel = self.rel
                # except: pass
        return newM

    def get(self, row_name, col_name):
        idxr = self.row_data.index(row_name)
        idxc = self.column_data.index(col_name)
        return self[idxr, idxc]

    def rows(self, component):
        rows = [self.row_data.index(item) for item in component]
        newM = self[rows,:]
        newM.model = self.model
        newM.column_data = self.column_data
        newM.row_data = list(np.array(self.row_data)[rows])
        return newM

    def del_zeros(self):
        delrows = np.where(~self.any(axis=1))[0]
        delcols = np.where(~self.any(axis=0))[0]

        newM = np.delete(self, delcols, axis=1).view(type(self))
        newM = np.delete(newM, delrows, axis=0).view(type(self))
        transfer_vars(self, newM)

        newM.column_data = list(np.delete(self.column_data, delcols))
        newM.row_data = list(np.delete(self.row_data, delrows))
        return newM
        
    def add_cols(self, component):
        if "colinear" in component:
            vertical = [elem for elem in self.model.elems if elem.Dx==0.0]
            other = set({})
            for elem in self.model.elems:
                try: other.add(elem.Dy/elem.Dx)
                except ZeroDivisionError: pass
            return vertical, other
        if type(component) is list:
            #ASSUMES component IS A LIST OF COLUMN INDICES
            delcols = [self.column_data.index(item) for item in component[1:len(component)]]
            i0 = self.column_data.index(component[0])
            newM = np.delete(self, delcols, axis=1).view(type(self))
            for col in delcols:
                newM[:,i0] += self[:,col]
            newM.row_data = self.row_data
            newM.model = self.model
            newM.column_data = list(np.delete(self.column_data, delcols))
            # try: newM.rel = self.rel
            # except: pass
            return newM

    def add_rows(self, component):
        if type(component) is list:
            delrows = [self.row_data.index(item) for item in component[1:len(component)]]
            i0 = self.row_data.index(component[0])
            newA = np.delete(self, delrows, axis=0).view(type(self))
            for row in delrows:
                newA[i0,:] += self[row,:]
            newA.column_data = self.column_data
            newA.model = self.model
            newA.row_data = list(np.delete(self.row_data, delrows))
            # try: newA.rel = self.rel
            # except: pass
            return newA


class nStatic_matrix(Structural_Matrix):
    ranges = {
        'f': 'free dof rows',
        'i': 'primary force columns',
        'x': 'redundant force columns',
        'c': 'continuous force columns (removes releases)',
        'd': 'reaction force columns',}
    def __new__(cls, arry, model, rcdata):
        return np.asarray(arry).view(cls)

    def __init__(self, arry, model, rcdata):
        self.model = model
        self.row_data = rcdata[0]
        self.column_data = rcdata[1]
        # if matrix is not None:
        #     fullnq = sum([len(elem.rel) for elem in model.elems])
        #     self[:,:] = np.zeros((model.nt,fullnq))
        #     for idxr, rw in enumerate(self.row_data):
        #         if rw in matrix.row_data:
        #             for idxc, cl in enumerate(self.column_data):
        #                 if cl in matrix.column_data:
        #                     self[idxr,idxc] = matrix.get(rw,cl)

    def __matmul__(self, Vector):
        if type(Vector) is P_vector:
            vect = np.matmul(self, Vector).view(Q_vector)
            vect.row_data = self.row_data
            vect.matrix = self
        elif type(Vector) is Q_vector:
            vect = np.matmul(self, Vector).view(P_vector) 
            vect.row_data = self.row_data
            vect.matrix = self
        else:
            vect = Structural_Matrix.__matmul__(self, Vector)
        return vect
  #<
    @property
    def f(self):
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) > self.model.nf]
        
        newB = np.delete(self, delrows, axis=0).view(type(self))
        transfer_vars(self, newB)
        # newB.model = self.model
        newB.row_data = list(np.delete(self.row_data, delrows))
        # newB.column_data = self.column_data
        return newB
        
    @property
    def i(self):
        """Removes rows of B_matrix corresponding to redundant forces"""
        # reducedB = self.f.c.del_zeros()
        reducedB = self.f.c
        rdts = reducedB.model.redundants
        tags = [q.elem.tag + '_'+str(q.nature) for q in rdts]
        delcols = [reducedB.column_data.index(tag) for tag in tags]
        newB = np.delete(reducedB, delcols, axis=1).view(Static_matrix)
        transfer_vars(reducedB, newB)
        newB.column_data = list(np.delete(reducedB.column_data, delcols))
        return newB
    
    @property
    def d(self):
        """Removes rows corresponding to free dofs"""
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) <= self.model.nf]
        newB = np.delete(self, delrows, axis=0).view(type(self))
        transfer_vars(self,newB)
        newB.row_data = list(np.delete(self.row_data, delrows))
        return newB

    @property
    def c(self):
        
        Bf = self.f
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delcols = [Bf.column_data.index(tag) for tag in tags]
        newB = np.delete(Bf, delcols, axis=1).view(type(self))
        transfer_vars(Bf, newB)
        newB.column_data = list(np.delete(Bf.column_data, delcols))
        # newB.row_data = self.row_data
        # newB.rel = self.rel
        return newB

    @property
    def o(self):
        """Remove columns corresponding to element force releases, then delete zeros"""
        Bf = self.f
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delcols = [Bf.column_data.index(tag) for tag in tags]
        newB = np.delete(Bf, delcols, axis=1).view(type(self))
        transfer_vars(Bf, newB)
        newB.column_data = list(np.delete(Bf.column_data, delcols))
        newB = newB.del_zeros()
        
        # newB.row_data = self.row_data
        # newB.rel = self.rel
        return newB

    @property
    def fc(self):
        return self.f.c
    
    @property
    def x(self):
        """Removes rows of B_matrix corresponding to primary (non-redundant) forces
        
        """
        rdts = self.model.redundants
        tags = [q.elem.tag + '_'+str(q.nature) for q in rdts]
        cols = [self.column_data.index(tag) for tag in tags]
        newB = self[:,cols]
        transfer_vars(self, newB)
        newB.column_data = [self.column_data[col] for col in cols]
        # newB.row_data = self.row_data
        # newB.model = self.model
        return newB

    @property
    def barxi(self):
        Bx = self.f.x
        Bbarxi = self.bari @ -Bx
        Bbarxi.column_data = Bx.column_data
        return Bbarxi

    @property
    def barx(self):
        # nQ = self.model.nQ
        nQ = len(self.column_data)
        nx = len(self.model.redundants)

        Bbarxi = self.barxi

        Bbarx = Structural_Matrix(np.zeros((nQ,nx)))
        transfer_vars(self, Bbarx)
        Bbarx.column_data = Bbarxi.column_data
        Bbarx.row_data = self.column_data

        # for idxr, rw in enumerate(Bbarx.row_data):
        #     if rw in Bbarxi.row_data:
        #         for idxc, cl in enumerate(Bbarx.column_data):
        #             print(rw,cl)
        #             if cl in Bbarxi.column_data:
        #                 Bbarx[idxr,idxc] = Bbarxi.get(rw,cl)
        #             elif cl==rw:
        #                 Bbarx[idxr, idxc] = 1.
        for idxc, cl in enumerate(Bbarx.column_data):
            for idxr, rw in enumerate(Bbarx.row_data):
                if rw in Bbarxi.row_data:
                    Bbarx[idxr,idxc] = Bbarxi.get(rw,cl)
                elif cl==rw:
                        Bbarx[idxr, idxc] = 1.

        return Bbarx

    @property
    def bari(self):
        return self.i.del_zeros().inv

    @property
    def c0(self):
        Bf = self.f
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delcols = [Bf.column_data.index(tag) for tag in tags]
        newB = Bf 
        for col in delcols:
            newB[:,col] = [0.]*len(Bf[:,0])
        transfer_vars(Bf, newB)
        newB.column_data = list(np.delete(Bf.column_data, delcols))
        # newB.row_data = self.row_data
        # newB.rel = self.rel
        return newB

class Kinematic_matrix(Structural_Matrix, np.ndarray):
    ranges = {
        'f': 'free dof columns',
        'i': 'primary force/deformation rows',
        'x': 'redundant force/deformation rows',
        'd': 'reaction force/deformation rows',
        'c': 'continuous (hingeless) force/deformation rows'
    }
    def __new__(cls, arry, model, rcitems):
        return np.asarray(arry).view(cls)

    def __init__(self, arry, model, rcitems):
        if rng is None:
            self.rng = None
        self.model = model
        self.column_data = [str(dof) for dof in range(1,model.nt+1)]
        # self.column_data = ['U_{'+str(dof)+'}' for dof in range(1, model.nt+1)]
        self.row_data = [elem.tag+'_'+key for elem in model.elems for key in elem.v.keys()]
        
        
        # self.rel = [rel for elem in model.elems for rel in elem.rel.values()]
        self.idx_h = []
        if matrix is not None:
            fullnq = sum([len(elem.rel) for elem in model.elems])
            self[:,:] = np.zeros((model.nt,fullnq))
            for idxr, rw in enumerate(self.row_data):
                if rw in matrix.row_data:
                    for idxc, cl in enumerate(self.column_data):
                        if cl in matrix.column_data:
                            self[idxr,idxc] = matrix.get(rw,cl)
  #<
    def __matmul__(self, Vector):
        if isinstance(Vector, Deformation_vector):
            # print('a')
            vect = np.matmul(self, Vector).view(Displacement_vector)
            vect.row_data = self.row_data
            vect.matrix = self
        elif isinstance(Vector, Displacement_vector):
            # print('b')
            vect = np.matmul(self, Vector).view(Deformation_vector)
            # vect = np.matmul(self, Vector).view(Structural_Vector)  
            vect.row_data = self.row_data
            vect.matrix = self
        else:
            # print('else')
            vect = Structural_Matrix.__matmul__(self, Vector)

        
        return vect

    def remove(self, component):
        if "frame-axial" in component:
            """Remove rows corresponding to frame axial deformations"""

            rows=[]
            for elem in self.model.elems:
                if type(elem)== emma.elements.Beam:
                    rows.append(self.row_data.index(elem.tag+'_1'))

            newA = np.delete(self, rows, axis=0).view(Kinematic_matrix)
            newA.row_data = list(np.delete(self.row_data, rows))
            newA.model = self.model
            newA.column_data = self.column_data
            # newA.rel = list(np.delete(self.rel, rows))
            return newA 

        if type(component) is list:
            newA = self
            for item in component:
                if item in self.column_data: 
                    delcol = self.column_data.index(item)
                    newA = np.delete(self, delcol, axis=1).view(Kinematic_matrix)
                    newA.row_data = self.row_data
                    newA.model = self.model
                    newA.column_data = list(np.delete(self.column_data, delcol))
                    # newA.rel = self.rel
                else:
                    delrow = self.row_data.index(item)
                    newA = np.delete(self, delrow, axis=0).view(Kinematic_matrix)
                    newA.column_data = self.column_data
                    newA.model = self.model
                    newA.row_data = list(np.delete(self.row_data, delrow))
                    # newA.rel = self.rel
            return newA

    def combine(self, component):
        if "colinear" in component:
            vertical = [elem for elem in self.model.elems if elem.Dx==0.0]
            other = set({})
            for elem in self.model.elems:
                try: other.add(elem.Dy/elem.Dx)
                except ZeroDivisionError: pass
            return vertical, other

        if type(component) is list:
            ## TO BE DEPRECATED
            #ASSUMES component IS A LIST OF COLUMN INDICES
            delcols = [self.column_data.index(item) for item in component[1:len(component)]]
            i0 = self.column_data.index(component[0])
            newA = np.delete(self, delcols, axis=1).view(Kinematic_matrix)
            for col in delcols:
                newA[:,i0] += self[:,col]
            newA.row_data = self.row_data
            newA.model = self.model
            newA.column_data = list(np.delete(self.column_data, delcols))
            # newA.rel = self.rel
            return newA
    
    @property
    def f(self):
        """Removes columns corresponding to fixed dofs"""
        delcols = [idx for idx, dof in enumerate(self.column_data) if int(dof) > self.model.nf]
        newA = np.delete(self, delcols, axis=1).view(Kinematic_matrix)
        newA.column_data = list(np.delete(self.column_data, delcols))
        newA.row_data = self.row_data
        newA.model = self.model
        # newA.rel = self.rel
        return newA

    @property
    def i(self):
        """Removes rows corresponding to redundant forces"""
        # tags = [q.elem.tag + str(q.nature) for q in rdts]
        # delrows = [self.row_data.index(tag) for tag in tags]
        # newA = np.delete(self, delrows, axis=0).view(Kinematic_matrix)
        # newA.row_data = list(np.delete(self.row_data, delrows))
        # newA.column_data = self.column_data
        # newA.model = self.model
        # newA.rel = list(np.delete(self.rel, delrows))
        Afc = self.f.c
        rdts = self.model.redundants
        tags = [q.elem.tag +'_'+ str(q.nature) for q in rdts]
        delrows = [Afc.row_data.index(tag) for tag in tags]
        newA = np.delete(Afc, delrows, axis=0).view(Kinematic_matrix)
        transfer_vars(Afc,newA)
        newA.row_data = list(np.delete(Afc.row_data, delrows))
        return newA

    @property
    def d(self):
        """Removes columns corresponding to free dofs"""
        delcols = [idx for idx, dof in enumerate(self.column_data) if int(dof) <= self.model.nf]
        newA = np.delete(self, delcols, axis=1).view(Kinematic_matrix)
        transfer_vars(self,newA)
        newA.column_data = list(np.delete(self.column_data, delcols))
        return newA

    @property
    def c(self):
        """Removes rows corresponding to element hinges/releases"""
        Af = self.f
        # delrows = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        delrows = [Af.row_data.index(tag) for tag in tags]
        newA = np.delete(Af, delrows, axis=0).view(Kinematic_matrix)
        transfer_vars(Af, newA)
        newA.row_data = list(np.delete(Af.row_data, delrows))
        # newA.column_data = Af.column_data
        # newA.model = Af.model
        # newA.rel = list(np.delete(Af.rel, delrows))
        return newA
    
    @property
    def c0(self):
        Af = self.f
        n_col = len(Af.T)
        tags = [elem.tag + "_" + rel for elem in self.model.elems for rel in elem.rel if elem.rel[rel]]
        # delcols = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        delrows = [Af.row_data.index(tag) for tag in tags]
        newA = Af 
        for rw in delrows:
            newA[rw,:] = [0.]*n_col
        transfer_vars(Af, newA)
        # newA.row_data = list(np.delete(Af.row_data, delrows))
        return newA

class Flexibility_matrix (Structural_Matrix):
    tag = 'F'
    def __new__(cls, arry, model, rcitems):
        return np.asarray(arry).view(cls)

    def __init__(self, arry, model, rcitems):
        # self.model = model 
        # self.column_data = [elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()]
        # self.row_data = [elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()]
        # self.rel = [rel for elem in model.elems for rel in elem.rel.values()]
        pass

    def __matmul__(self, Vector):
        if type(Vector) is Deformation_vector:
            vect = np.matmul(self, Vector).view(Q_vector)
            vect.row_data = self.row_data
            vect.matrix = self
        elif type(Vector) is Q_vector:
            vect = np.matmul(self, Vector).view(Deformation_vector)
            # vect.part = 'continuous'
            vect.row_data = self.row_data
            vect.matrix = self

        else:
            vect = Structural_Matrix.__matmul__(self, Vector)

        return vect
  #<
    @property 
    def c(self):
        """Removes rows corresponding to element hinges/releases"""
        rows = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        newF = np.delete(self, rows, axis=0).view(type(self))
        newF = np.delete(newF, rows, axis=1).view(type(self))
        newF.row_data = list(np.delete(self.row_data, rows))
        newF.column_data = list(np.delete(self.column_data, rows))
        newF.model = self.model
        newF.tag = self.tag + "_c"
        return newF

class Stiffness_matrix (Structural_Matrix):
    tag = 'K'
    def __new__(cls, arry, model, rcitems):
        return np.asarray(arry).view(cls)

    def __init__(self, arry, model, rcitems):
        self.subs = [None]
        pass
    
    def __matmul__(self, Vector):
        if type(Vector) is Displacement_vector:
            vect = np.matmul(self, Vector).view(P_vector)
            vect.row_data = self.row_data
            vect.matrix = self
        elif type(Vector) is P_vector:
            vect = np.matmul(self, Vector).view(Displacement_vector)
            vect.row_data = self.row_data
            vect.matrix = self
        else:
            vect = Structural_Matrix.__matmul__(self, Vector)
        return vect
  #<
    @property
    def f(self):
        delrows = [idx for idx, dof in enumerate([str(dof) for dof in range(1, self.model.nt+1)]) if int(dof) > self.model.nf]
        
        newK = np.delete(self, delrows, axis=0)
        newK = np.delete(newK, delrows, axis=1).view(type(self))
        transfer_vars(self, newK)
        
        newK.row_data = list(np.delete(self.row_data, delrows))
        newK.column_data = list(np.delete(self.column_data, delrows))
        return newK

class Mass_matrix(Structural_Matrix):
    tag = 'M'
    def __new__(cls, arry, model, rcitems):
        return np.asarray(arry).view(cls)
  #<
    @property
    def f(self):
        delrows = [idx for idx, dof in enumerate([str(dof) for dof in range(1, self.model.nt+1)]) if int(dof) > self.model.nf]
        
        newK = np.delete(self, delrows, axis=0)
        newK = np.delete(newK, delrows, axis=1).view(type(self))
        transfer_vars(self, newK)
        
        newK.row_data = list(np.delete(self.row_data, delrows))
        newK.column_data = list(np.delete(self.column_data, delrows))
        return newK
    
    @property
    def m(self):
        delrows = [idx for idx, dof in enumerate([str(dof) for dof in range(1, self.model.nt+1)]) if int(dof) > self.model.nf]
        
        newK = np.delete(self, delrows, axis=0)
        newK = np.delete(newK, delrows, axis=1).view(type(self))
        transfer_vars(self, newK)
        
        newK.row_data = list(np.delete(self.row_data, delrows))
        newK.column_data = list(np.delete(self.column_data, delrows))
        return newK


class Deformation_vector(Structural_Vector):
    tag = "V"
    def __new__(cls, arry, model, Vector=None):
        if part == 'initial':
            V = np.concatenate([elem.v0_vector() for elem in Matrix.model.elems])
        else:
            V = np.zeros((len(Matrix.row_data)))
        input_array = np.asarray(V).view(cls)
        input_array.matrix = Matrix
        return input_array

    def __init__(self, arry, model, Vector=None):
        self.part = part 
        self.subs = [None]
        if part is None: 
            self.o = Deformation_vector(Matrix, part='initial')
            self.o.subs.append('initial')
            pass
        self.row_data = Matrix.row_data
        
        if Vector is not None:
            for key in Vector.row_data:
                if key in self.row_data:
                    self.set_item(key, Vector.rows([key]))
  #<    
    @property 
    def c(self):
        """Removes rows corresponding to element hinges/releases"""
        # rows = [idx for idx, rel in enumerate(self.matrix.model.rel) if rel]
        tags = [elem.tag + "_" + rel for elem in self.matrix.model.elems for rel in elem.rel if elem.rel[rel]]
        delrows = [self.row_data.index(tag) for tag in tags]
        
        newV = np.delete(self, delrows, axis=0).view(type(self))
        transfer_vars(self,newV)
        newV.row_data = list(np.delete(self.row_data, delrows))
        newV.subs.append('continuous')
        return newV  
    
    @property
    def i(self):
        """Removes rows corresponding to redundant forces"""
        rdts = self.matrix.model.redundants
        tags = [q.elem.tag + '_'+str(q.nature) for q in rdts]
        delrows = [self.row_data.index(tag) for tag in tags]
        newV = np.delete(self, delrows, axis=0).view(type(self))
        transfer_vars(self,newV)
        newV.row_data = list(np.delete(self.row_data, delrows))
        newV.subs.append('primary')
        return newV

    @property
    def x(self):
        """Removes rows of corresponding to primary forces
        
        """
        rdts = self.matrix.model.redundants
        tags = [q.elem.tag + '_'+str(q.nature) for q in rdts]
        rows = [self.row_data.index(tag) for tag in tags]
        newV = self[rows]
        newV.row_data = [self.row_data[row] for row in rows]
        transfer_vars(self, newV)
        return newV
        
class Load_Vector(Structural_Vector):
    tag = "P"
    def __new__(cls, arry, model, Vector=None):
        P = np.zeros(Static_matrix.shape[0])
        if part is None:
            for node in Static_matrix.model.nodes:
                p = node.p_vector()
                for i, dof in enumerate(node.dofs):
                    P[int(dof)-1] += p[i]
            # P = np.concatenate([node.p_vector() for node in Static_matrix.model.nodes])
       
        elif part =='elem-load':
            for elem in Static_matrix.model.elems:
                if type(elem) is emma.Beam:
                    dofs = elem.dofs
                    P[int(dofs[0])-1] +=  elem.w['y']*elem.L/2*elem.sn
                    P[int(dofs[1])-1] += -elem.w['y']*elem.L/2*elem.cs
                    P[int(dofs[3])-1] +=  elem.w['y']*elem.L/2*elem.sn
                    P[int(dofs[4])-1] += -elem.w['y']*elem.L/2*elem.cs
                
        return np.asarray(P).view(cls)

    def __init__(self, arry, model, Vector=None):
        self.model = model
        self.subs = [None]
        self.row_data = Static_matrix.row_data
  #<
    @property
    def f(self):
        """Removes rows corresponding to fixed dofs"""
        
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) > self.matrix.model.nf]
        newP = np.delete(self, delrows, axis=0).view(type(self))
        newP.row_data = list(np.delete(self.row_data, delrows))
        newP.model = self.model
        newP.subs.append('free')
        return newP
    
    @property
    def wf(self):
        return self.w.f

    @property
    def d(self):
        """Removes rows corresponding to free dofs"""
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) <= self.matrix.model.nf]
        newP = np.delete(self, delrows, axis=0).view(type(self))
        newP.model = self.model
        newP.row_data = list(np.delete(self.row_data, delrows))
        newP.subs.append('reactions')
        return newP

class Pw_vector(Structural_Vector):
    tag = "P_w"
    def __new__(cls, model):
        P = np.zeros(model.nt)
        for elem in model.elems:
            if type(elem) is emma.Beam:
                dofs = elem.dofs
                P[int(dofs[0])-1] +=  elem.w['y']*elem.L/2*elem.sn
                P[int(dofs[1])-1] += -elem.w['y']*elem.L/2*elem.cs
                P[int(dofs[3])-1] +=  elem.w['y']*elem.L/2*elem.sn
                P[int(dofs[4])-1] += -elem.w['y']*elem.L/2*elem.cs
        return np.asarray(P).view(cls)

    def __init__(self, model, part=None):
        self.model = model
        self.subs = [None]
        if part == None:
            self.w = P_vector(Static_matrix, part='elem-load')
            self.w.subs.append('elem-load')
        self.row_data = [str(dof) for dof in range(1, model.nt+1)]
        
    @property
    def f(self):
        """Removes rows corresponding to fixed dofs"""
        
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) > self.model.nf]
        newP = np.delete(self, delrows, axis=0).view(type(self))
        newP.row_data = list(np.delete(self.row_data, delrows))
        newP.model = self.model
        newP.subs.append('free')
        return newP

    @property
    def d(self):
        """Removes rows corresponding to free dofs"""
        delrows = [idx for idx, dof in enumerate(self.row_data) if int(dof) <= self.matrix.model.nf]
        newP = np.delete(self, delrows, axis=0).view(type(self))
        newP.model = self.model
        newP.row_data = list(np.delete(self.row_data, delrows))
        newP.subs.append('reactions')
        return newP

class Displacement_vector(Structural_Vector):
    tag = 'U'
    def __new__(cls, arry, model, Vector=None):
        return np.asarray(arry).view(cls)

    def __init__(self, arry, model, Vector=None):
        self.model = model 
        self.row_data = Kinematic_matrix.column_data
        self.subs = [None]
        if Vector is not None:
            for key in Vector.row_data:
                if key in self.row_data:
                    self.set_item(key, Vector.rows([key]))
  #<
    @property
    def f(self):
        """Removes rows corresponding to fixed dofs"""
        delrows = [idx for idx, dof in enumerate(self.matrix.column_data) if int(dof) > self.matrix.model.nf]
        newU = np.delete(self, delrows, axis=0).view(Displacement_vector)
        newU.row_data = list(np.delete(self.row_data, delrows))
        newU.model = self.model
        return newU

class iForce_vector(Structural_Vector):
    tag = 'Q'
    def __new__(cls, arry, model, row_data, Vector=None):
        return np.asarray(arry).view(cls)

    def __init__(self, arry, model, row_data, Vector=None):
        self.model = model
        self.subs = [None]
        self.row_data = row_data
        if Vector is not None:
            for key in Vector.row_data:
                if key in self.row_data:
                    self.set_item(key, Vector.get(key))
  #<
    @property
    def i(self):
        """Removes rows corresponding to redundant forces"""
        rdts = self.model.redundants

        tags = [q.elem.tag + str(q.nature) for q in rdts]
        delrows = [self.row_data.index(tag) for tag in tags]
        newQ = np.delete(self, delrows, axis=0).view(iForce_Vector)
        transfer_vars(self, newQ)
        newQ.subs.append('primary')
        newQ.row_data = list(np.delete(self.row_data, delrows))
        return newQ
    
    @property 
    def x(self):
        pass

class DiagFlex_matrix (Structural_Matrix):
    """FS_MATRIX block diagonal matrix of element flexibity matrices for structural model
    FS = FS_MATRIX (MODEL,ELEMDATA,ROPTION)
    the function sets up the block diagonal matrix of element flexibility matrices FS
    for the structural model specified in data structure MODEL with element property
    information in cell array ELEMDATA;
    if ROPTION=0, element release information is not accounted for in setting up Fs (default=1)
    
    Parameters
    =========================================================================================
    model
    
    -----------------------------------------------------------------------------------------
    """
    tag = 'F'
    def __new__(cls, model, Roption=None):
        return np.asarray(arry).view(cls)

    def __init__(self, model, Roption=None):
        # self.model = model 
        # self.column_data = [elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()]
        # self.row_data = [elem.tag+'_'+key for elem in model.elems for key in elem.rel.keys()]
        # self.rel = [rel for elem in model.elems for rel in elem.rel.values()]
        pass

    def __matmul__(self, Vector):
        if type(Vector) is Deformation_vector:
            vect = np.matmul(self, Vector).view(Q_vector)
            vect.row_data = self.row_data
            vect.matrix = self
        elif type(Vector) is Q_vector:
            vect = np.matmul(self, Vector).view(Deformation_vector)
            # vect.part = 'continuous'
            vect.row_data = self.row_data
            vect.matrix = self

        else:
            vect = Structural_Matrix.__matmul__(self, Vector)
        return vect

    @property 
    def c(self):
        """Removes rows corresponding to element hinges/releases"""
        rows = [idx for idx, rel in enumerate(self.model.rel) if rel==1]
        newF = np.delete(self, rows, axis=0).view(type(self))
        newF = np.delete(newF, rows, axis=1).view(type(self))
        newF.row_data = list(np.delete(self.row_data, rows))
        newF.column_data = list(np.delete(self.column_data, rows))
        newF.model = self.model
        newF.tag = self.tag + "_c"
        return newF

class DiagStif_matrix (Structural_Matrix):
    """...
    Parameters
    =========================================================================================
    model
    
    -----------------------------------------------------------------------------------------
    """    
    tag = 'K'
    def __new__(cls, arry, model, Roption=None):
        return np.asarray(arry).view(cls)

    def __init__(self, model, Roption=None):
        self.subs = [None]
        pass



import tensorflow as tf


class Inputs(object):
    def __init__(self, img1: tf.Tensor, img2: tf.Tensor, label: tf.Tensor):
        self.img1 = img1
        self.img2 = img2
        self.label = label

class Truss(Inputs):
    def __init__(self,E,A,L):
        pass

class ElemSet(object):
    img1_resized = 'img1_resized'
    img2_resized = 'img2_resized'
    label = 'same_person'

    def __init__(self, generator=PairGenerator()):
        self.next_element = self.build_iterator(generator)

    def build_iterator(self, pair_gen: PairGenerator):
        batch_size = 10
        prefetch_batch_buffer = 5

        dataset = tf.data.Dataset.from_generator(pair_gen.get_next_pair,
                                                 output_types={PairGenerator.person1: tf.string,
                                                               PairGenerator.person2: tf.string,
                                                               PairGenerator.label: tf.bool})
        dataset = dataset.map(self._read_image_and_resize)
        dataset = dataset.batch(batch_size)
        dataset = dataset.prefetch(prefetch_batch_buffer)
        iter = dataset.make_one_shot_iterator()
        element = iter.get_next()

        return Inputs(element[self.img1_resized],
                      element[self.img2_resized],
                      element[PairGenerator.label])

    def _read_image_and_resize(self, pair_element):
        target_size = [128, 128]
        # read images from disk
        img1_file = tf.read_file(pair_element[PairGenerator.person1])
        img2_file = tf.read_file(pair_element[PairGenerator.person2])
        img1 = tf.image.decode_image(img1_file)
        img2 = tf.image.decode_image(img2_file)

        # let tensorflow know that the loaded images have unknown dimensions, and 3 color channels (rgb)
        img1.set_shape([None, None, 3])
        img2.set_shape([None, None, 3])

        # resize to model input size
        img1_resized = tf.image.resize_images(img1, target_size)
        img2_resized = tf.image.resize_images(img2, target_size)

        pair_element[self.img1_resized] = img1_resized
        pair_element[self.img2_resized] = img2_resized
        pair_element[self.label] = tf.cast(pair_element[PairGenerator.label], tf.float32)

        return pair_element


class Model(object):
    def __init__(self, inputs: Inputs):
        self.inputs = inputs
        self.predictions = self.predict(inputs)
        self.loss = self.calculate_loss(inputs, self.predictions)
        self.opt_step = tf.train.AdamOptimizer(learning_rate=0.001).minimize(self.loss)

    def predict(self, inputs: Inputs):
        with tf.name_scope("image_substraction"):
            img_diff = (inputs.img1 - inputs.img2)
            x = img_diff
        with tf.name_scope('conv_relu_maxpool'):
            for conv_layer_i in range(5):
                x = tf.layers.conv2d(x,
                                     filters=20 * (conv_layer_i + 1),
                                     kernel_size=3,
                                     activation=tf.nn.relu)
                x = tf.layers.max_pooling2d(x,
                                            pool_size=3,
                                            strides=2)
        with tf.name_scope('fully_connected'):
            for conv_layer_i in range(1):
                x = tf.layers.dense(x,
                                    units=200,
                                    activation=tf.nn.relu)
        with tf.name_scope('linear_predict'):
            predicted_logits = tf.layers.dense(x, 1, activation=None)

        return tf.squeeze(predicted_logits)

    def calculate_loss(self, inputs: Inputs, prediction_logits: Tensor):
        with tf.name_scope('calculate_loss'):
            return tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(labels=inputs.label,
                                                                          logits=prediction_logits))




import numpy as np
import pandas as pd
# import sympy as sp
from . import export
import scipy.linalg
from . import ipyutils
from . import numerical

settings = {
    "DATAFRAME_LATEX": True, 
}

subscripts = {
    'part': 'subscript',
     None: '',
    'initial': '0',
    'continuous':'c',
    'primary':'i',
    'reactions':'d',
    'elem-load':'w',
    'free': 'f',
}

matmul = {
    'B':{'Q':'P'}
}



def del_zeros(mat):
    delrows = np.where(~mat.any(axis=1))[0]
    delcols = np.where(~mat.any(axis=0))[0]

    newM = np.delete(mat, delcols, axis=1)
    newM = np.delete(newM, delrows, axis=0)
    return newM

def transfer_vars(item1, item2):
    for key, value in item1.__dict__.items():
        item2.__dict__[key] = value

class Structural_Vector(np.ndarray):
    column_data = ["vector"]
    row_data = None
    subs = [None]
    tag = 'Vector'
    def __new__(cls, mat):
        mat = mat
        return np.asarray(mat).view(cls)

    def __init__(self, mat):
        self.tag = None

    def _repr_html_(self):
        try:
            out = self.df.to_html()
            return out
        except: 
            return self

    def __add__(self, other):
        if isinstance(other, type(self)):
            out = np.add(self, other).view(type(self))
            transfer_vars(self,out)
        else:
            out = super().__add__(other)
        return out
    
    def get(self, key):
        idx = np.array([i for i,j in enumerate(self.row_data) if str(j) == key], dtype=int)
        row = self.row_data.index(component)
        return self[idx]

    def set_item(self, key, value):
        idx = np.array([i for i,j in enumerate(self.row_data) if str(j) == key], dtype=int)
        self[idx] = value

    def rows(self, component):
        rows = [self.row_data.index(item) for item in component]
        idx = np.array([i for i,j in enumerate(self.row_data) if str(j) == key])
        newV = self[idx]
        newV.row_data = list(np.array(self.row_data)[idx])
        newV.model = self.model
        return newV

    @property
    def df(self):
        row_data = ['$'+str(tag)+'$' for tag in self.row_data]
        header = '$'+self.tag+'_{{'
        try:
            for sub in self.subs:
                header += subscripts[sub]
        except: pass
        print
        header += '}}$'

        return pd.DataFrame(np.around(self,14), index=row_data, columns=[header])

    @property
    def symb(self):
        var = []
        for eid in self.row_data:
            var.append(sp.symbols(self.tag+eid))
        return sp.Matrix(var)
    
    @property
    def disp(self):
        return sp.Matrix(self)
    
    
class Structural_Matrix(np.ndarray):
    column_data = None
    row_data = None
    c_ridx = None # row indexes for use in .c method
    c_cidx = None # column indexes for use in .c method
    tag = None

    def __new__(cls, mat):
        mat = mat
        return np.asarray(mat).view(cls)

    def __init__(self, mat):
        self.tag = None

    def __matmul__(self, other):
        if isinstance(other, Structural_Matrix):
            out = np.matmul(self,other).view(Structural_Matrix)
            transfer_vars(self,out)
            out.column_data = other.column_data
        
        elif isinstance(other, Structural_Vector):
            out = np.matmul(self,other).view(Structural_Vector)
            out.row_data = self.row_data
            # out.column_data = ['.']
        else:
            out = np.matmul(self,other).view(Structural_Vector)
            out.row_data = self.row_data
            # out.column_data = ['.']
        return out

    def __add__(self, other):
        out = np.add(self,other)
        if (isinstance(other, Structural_Matrix)
            or isinstance(other, Structural_Vector)):
            out.row_data = self.row_data

        return out
    
    def __truediv__(self, other):
        out = np.ndarray.__truediv__(self, other) 
        if (isinstance(other, float) 
            or isinstance(other, Structural_Matrix)
            or isinstance(other, Structural_Vector)):
            out = np.ndarray.__truediv__(self, other).view(Structural_Matrix)
            transfer_vars(self, out)
        else:
            out = np.ndarray.__truediv__(self, other)
        return out
        
    def _repr_html_(self):
        try: 
            df = self.df
            return df.to_html()
        except:
            try:
                return pd.DataFrame(self).to_html()
            except:
                pass

    @property
    def disp(self):
        return sp.Matrix(self)
    
    @property
    def df(self):
        if settings['DATAFRAME_LATEX']:
            row_data = ['$'+str(tag)+'$' for tag in self.row_data]
            column_data = ['$'+str(tag)+'$' for tag in self.column_data]
        else:
            row_data = [str(i) for i in self.row_data]
            column_data = [str(i) for i in self.column_data]
        return pd.DataFrame(np.around(self,14), index=row_data, columns=column_data)

    @property
    def inv(self):
        mat = np.linalg.inv(self)
        transfer_vars(self, mat)
        mat.row_data = self.column_data
        mat.column_data = self.row_data
        return mat
    

    @property
    def rank(self):
        """Return the rank of a matrix"""
        return np.linalg.matrix_rank(self)
    
    @property
    def lns(self):
        """Return a basis for the left nullspace of a matrix."""
        return scipy.linalg.null_space(self.T)
    
    @property
    def nls(self):
        """return a basis for the nullspace of matrix."""
        return scipy.linalg.null_space(self)
    
    @property
    def ker(self):
        "Return a basis for the kernel (nullspace) of a matrix."
        kernel = scipy.linalg.null_space(self) 
        ker = Structural_Matrix(kernel)
        ker.row_data = self.column_data 
        ker.column_data = [str(i+1) for i in range(len(ker[0]))]
        ker.model = self.model
        return ker

    def lu(self):
        
        return scipy.linalg.lu(self)


    @property
    def c(self):
        delcols = self.c_cidx
        delrows = self.c_ridx

        newM = np.delete(self, delrows, axis=0).view(type(self))
        if delcols: newM = np.delete(newM, delcols, axis=1).view(type(self))
        
        transfer_vars(self, newM)

        if delcols: newM.column_data = list(np.delete(self.column_data, delcols))

        newM.row_data = list(np.delete(self.row_data, delrows))
        return newM

    def round(self, num):
        newM = np.around(self, num).view(Structural_Matrix)
        transfer_vars(self,newM)
        return newM

    def remove(self, component):
        """Remove items by looking up column_data/row_data"""
        if type(component) is list:
            for item in component:
                if item in self.column_data: 
                    delcol = self.column_data.index(item)
                    newM = np.delete(self, delcol, axis=1).view(type(self))
                    transfer_vars(self,newM)
                    newM.column_data = list(np.delete(self.column_data, delcol))
                else:
                    delrow = self.row_data.index(item)
                    newM = np.delete(self, delrow, axis=0).view(type(self))
                    transfer_vars(self, newM)
                    # newM.column_data = self.column_data
                    # newM.model = self.model
                    newM.row_data = list(np.delete(self.row_data, delrow))

        else:
            item = component
            if item in self.column_data: 
                delcol = self.column_data.index(item)
                newM = np.delete(self, delcol, axis=1).view(type(self))
                newM.row_data = self.row_data
                newM.model = self.model
                newM.column_data = list(np.delete(self.column_data, delcol))
                # try: newM.rel = self.rel
                # except: pass
            else:
                delrow = self.row_data.index(item)
                newM = np.delete(self, delrow, axis=0).view(type(self))
                newM.column_data = self.column_data
                newM.model = self.model
                newM.row_data = list(np.delete(self.row_data, delrow))
                # try: newM.rel = self.rel
                # except: pass
        return newM

    def get(self, row_name, col_name):
        idxr = np.where(self.row_data == row_name)
        idxc = np.where(self.column_data == col_name)
        return self[idxr, idxc]


    def rows(self, component):
        rows = [self.row_data.index(item) for item in component]
        newM = self[rows,:]
        newM.model = self.model
        newM.column_data = self.column_data
        newM.row_data = list(np.array(self.row_data)[rows])
        return newM

    def del_zeros(self):
        """Delete rows and columns of a matrix with all zeros"""
        delrows = np.where(~self.any(axis=1))[0]
        delcols = np.where(~self.any(axis=0))[0]

        newM = np.delete(self, delcols, axis=1).view(type(self))
        newM = np.delete(newM, delrows, axis=0).view(type(self))
        transfer_vars(self, newM)

        newM.column_data = list(np.delete(self.column_data, delcols))
        newM.row_data = list(np.delete(self.row_data, delrows))
        return newM
        

    def add_cols(self, component):
        if "colinear" in component:
            vertical = [elem for elem in self.model.elems if elem.Dx==0.0]
            other = set({})
            for elem in self.model.elems:
                try: other.add(elem.Dy/elem.Dx)
                except ZeroDivisionError: pass
            return vertical, other
        if type(component) is list:
            #ASSUMES component IS A LIST OF COLUMN INDICES
            delcols = [self.column_data.index(item) for item in component[1:len(component)]]
            i0 = self.column_data.index(component[0])
            newM = np.delete(self, delcols, axis=1).view(type(self))
            for col in delcols:
                newM[:,i0] += self[:,col]
            newM.row_data = self.row_data
            newM.model = self.model
            newM.column_data = list(np.delete(self.column_data, delcols))
            # try: newM.rel = self.rel
            # except: pass
            return newM

    def add_rows(self, component):
        if type(component) is list:
            delrows = [self.row_data.index(item) for item in component[1:len(component)]]
            i0 = self.row_data.index(component[0])
            newA = np.delete(self, delrows, axis=0).view(type(self))
            for row in delrows:
                newA[i0,:] += self[row,:]
            newA.column_data = self.column_data
            newA.model = self.model
            newA.row_data = list(np.delete(self.row_data, delrows))
            return newA

